/***********************************************************************
 * File: abacus.c
 * 	(Extracted from
 *	   "$Id: abacus.c,v 1.5 2009/06/22 15:09:44 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source file implements `abacus()' API.
 *
 * Copyright (C) 2006-2009 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#include "abacus.h"
#if !defined(LLONG_MAX) /* { cause of `splint' doesn't know LLONG_MAX */
#	define LLONG_MAX	((long long)0x7fffffffffffffff)
#endif /* } !defined(LLONG_MAX) */

char	*register_0 = "r0"; /* specifier of register#0 */

/* table to convert an integer digit to the character code */
static char i2a[] =
		{ '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' };

/* table of the exponents of ten */
static unsigned long long ten_E[MAX_Digits + 1] = {
	  1ULL				/* ten_E[0] = 10E0 */
	, 10ULL
	, 100ULL
	, 1000ULL
	, 10000ULL
	, 100000ULL
	, 1000000ULL
	, 10000000ULL
	, 100000000ULL
	, 1000000000ULL
	, 10000000000ULL
	, 100000000000ULL
	, 1000000000000ULL
	, 10000000000000ULL
	, 100000000000000ULL
	, 1000000000000000ULL
	, 10000000000000000ULL
	, 100000000000000000ULL
	, 1000000000000000000ULL	/* ten_E[18] = 10E18 */
	};
#define MAX_Exp			MAX_Digits
#define MAX_Int_part	(ten_E[MAX_Exp] - 1)
#define MAX_Fra_part	MAX_Int_part
#define ELEM_N_Multiply_vector										\
						(4) /* XXX (2 * (sizeof(ten_E[0]) / sizeof(unsigned long))) */
#define SZ_Numerator	(3 * MAX_Digits + 1)


/* calculating data structure */
typedef struct data_t { /* data type of this machine */
	int					is_negative;
	#define				IS_Negative(D) (((D)->is_negative != 0)? 1: 0)
	unsigned long long	int_part, fra_part;
} Data_t;
#define IS_Zero(dp) (((dp)->int_part == 0ULL) && (dp)->fra_part == 0ULL)


/* abacus architecture: register#0, a program-counter,
 *			the working memory allocation control
 *			and counter of commands
 */
typedef struct machine_t { /* definition of this machine */
	Data_t	r0;		/* register of calculating results */
	size_t	pc;		/* program counter */
	size_t	memories_n;	/* elements numbers of memories[] */
	Data_t	*memories;	/* pointing to working memories */
	size_t	commands_n;	/* number of invoker's command list */
} Machine_t;

/* predication: the specified step is illegal, or not */
#define IS_Illegal_skip_steps(STEPS,MAC) (((						\
	(STEPS) == 0LL)													\
	|| (((STEPS) < 0LL)												\
			&& (((STEPS) == (ssize_t)LONG_MIN)						\
				|| (ssize_t)((STEPS) * -1) > (ssize_t)((MAC)->pc)	\
				)													\
			)														\
		)															\
	|| (((STEPS) > 0ll)												\
			&& (STEPS) >= (ssize_t)((MAC)->commands_n - (MAC)->pc)	\
			)														\
	)



/* create the abacus architecture */
static Machine_t *create_machine(Machine_t	*mac
								, size_t	memories_n
								, size_t	commands_n
								)
{
	/* test: arguments are invalid? */
	if (mac == NULL) {
		/* ERROR: invalid arguments */
		errno = EINVAL; return (Machine_t *)NULL;
	}

	/* initialize the register#0 and program counter `pc' */
	memset(&(mac->r0), (int)'\0', sizeof(*mac)); mac->pc = 0;

	/* test: invoker needs the working memory? */
	if ((memories_n > 0) && mac->memories == NULL) {
		/* test: failed to allocate the working memories */
		if ((mac->memories = (Data_t *)malloc(
									sizeof(mac->memories[0]) * memories_n
									)) == NULL
		) {	/* ERROR: insufficient memory */
			errno = ENOMEM; return (Machine_t *)NULL;
		}

		/* initialize the register#0 and working memories */
		memset(mac->memories
				, (int)'\0'
				, (sizeof(mac->memories[0]) * memories_n)
				);
	}
	else {	/* invoker doesn't need working memory */
		mac->memories	= (Data_t *)NULL;
	}

	/* set the number of working memory slots */
	mac->memories_n	= memories_n;

	/* set the number of invoker's command list */
	mac->commands_n	= commands_n;

	return	mac; /* normal end */
}


/* release abacus architecture */
static Machine_t *destroy_machine(Machine_t *mac)
{
	/* test: arguments are invalid? */
	if (mac == NULL) {
		/* ERROR: invalid arguments */
		errno = EINVAL; return (Machine_t *)NULL;
	}

	/* test: working memory was allocated? */
	if ((mac->memories_n > 0) && mac->memories != NULL) {
		/* release the working memories */
		free(memset(mac->memories
					, (int)'\0'
					, (sizeof(mac->memories[0]) * mac->memories_n)
					)
			);
		mac->memories	= NULL;
	}

	/* clean up the machine */
	mac->memories_n	= 0;
	memset(&(mac->r0), (int)'\0', sizeof(mac->r0));
	mac->pc			= 0;
	mac->commands_n	= 0;

	return	mac; /* normal end */
}


/* get the address of memory-object that is specified operand */
static void *get_address(Machine_t *mac, void *operand)
{
	void	*address;

	/* test: operand indicates register#0? */
	if (operand == (void *)register_0) {
		address	= (void *)(&(mac->r0)); /* set `r0' address */
	}
	else if ((size_t)operand < mac->memories_n) {
		/* operand indicates slot number of working memories */
		address	= (void *)(mac->memories + (size_t)operand);
	}
	else {	/* operand points to invoker's string buffer */
		address	= operand;
	}

	return	address; /* always, normal end */
}


/* retrieve the `Value' operand string */
static char *retrieve_operand(char *operand, char *buf)
{
	size_t	len = strlen(operand);

	/* test: the length of operand is greater than maximum? */
	if (len > SZ_Numeral_str - 1) {
		len	= SZ_Numeral_str - 1; /* set length maximum */
	}

	/* transfer the operand string to destiny buffer */
	memcpy(memset(buf, (int)'\0', len + 1), operand, len); buf[len] = '\0';

	return	buf; /* always, normal end */
}


/* convert a string to the calculating data structure */
static Data_t *str_to_data(char *str, Data_t *data)
{
	size_t	fra_len;
	char	*endp, *p = str;

	/* determine the signature */
	for (data->is_negative = 0; *p != '\0'; ++p) {
		if ((*p == '+') || isspace(*p)) {	continue;	}
		if ((*p == '.') || isdigit(*p)) {	break;		}
		if (*p == '-') { /* found negative signature */
			data->is_negative	= !0;
			break;
		}
		else {	/* ERROR: invalid arguments */
			errno = EINVAL; return (Data_t *)NULL;
		}
	}
	str	= (*p == '-')? p + 1: p; /* progress the source ptr */

	/* set integer part */
	if (*str == '.') { /* there isn't integer digits */
		data->int_part	= 0ULL;
		++str;
	}
	else {	/* parse the integer parts */
		char	sv_ch;

		/* search the decimal point '.' */
		if ((p = strchr(str, '.')) == NULL) {
			/* there isn't decimal point */
			p	= str + strlen(str);
		}

		/* test: integer digits is too long? */
		if ((p - str) > MAX_Digits) {
			/* ERROR: out of domain */
			errno = EDOM; return (Data_t *)NULL;
		}

		/* parse the integer digits */
		errno = 0; sv_ch = *p; *p = '\0'; endp = str;
		data->int_part = (unsigned long long)strtoll((const char *)str
													, &endp
													, 10
													);
		*p = sv_ch;
		if ((errno != 0) || endp != p) {
			/* ERROR: invalid arguments */
			errno = EINVAL; return (Data_t *)NULL;
		}

		/* progress the source pointer */
		str	= (sv_ch == '.')? p + 1: p;
	}

	/* set fractional part */
	if (*str == '\0') { /* there isn't the fractional digits */
		data->fra_part	= 0ULL;
	}
	else if ((fra_len = strlen(str)) > MAX_Digits) {
		/* fractional digits are too long => ERROR: out of domain */
		errno = EDOM; return (Data_t *)NULL;
	}
	else {	/* parse the fractional digits */
		errno = 0; endp = str;
		data->fra_part = (unsigned long long)strtoll((const char *)str
													, &endp
													, 10
													);
		if ((errno != 0) || endp != (str + strlen(str))) {
			/* ERROR: invalid arguments */
			errno = EINVAL; return (Data_t *)NULL;
		}

		/* shift up the fractional parts */
		data->fra_part	*= ten_E[MAX_Digits - fra_len];
	}

	/* test: the value is zero? */
	if (IS_Zero(data)) {
		data->is_negative = 0; /* zero is positive, always */
	}

	return	data; /* normal end */
}


/* convert a operand to the calculating data structure */
static Data_t *operand_to_data(Machine_t	*mac
								, void		*operand
								, Data_t	*data
								)
{
	char	buf[SZ_Numeral_str];
	Data_t	*address;

	/* test: the operand points to the invoker's string buffer? */
	if ((address = (Data_t *)get_address(mac, operand)) == operand) {
		/* test: the invoker's buffer is illegal?
		 *	(if it is legal, translate it to machine's data)
		 */
		if (str_to_data(retrieve_operand((char *)operand
										, (char *)memset(buf
														, (int)'\0'
														, sizeof(buf)
														)
										)
						, data
						) == NULL
		) { return	(Data_t *)NULL; /* ERROR: occurred several errors */ }
	}
	else {	/* copy the register#0 or working memory */
		memcpy(data, address, sizeof(*data));
	}

	return	data; /* normal end */
}


/* convert a calculating data structure to the string */
static char *data_to_str(Data_t		*data
						, char		*str
						, size_t	int_digits
						, size_t	fra_digits
						)
{
	/* test: data is out of range? */
	if ((data->int_part >= ten_E[int_digits])
	|| data->fra_part > MAX_Fra_part
	) { errno = ERANGE; return (char *)NULL; /* ERROR: out of range */ }

	/* convert data to string */
	if (fra_digits != 0) { /* fractional digits are needed */
		if (int_digits != 0) {
			/* integer and fractional digits are needed */
			(void)snprintf(str
						, SZ_Numeral_str
						, "%s%llu.%0*llu"
						, ((IS_Negative(data) != 0)? "-": "")
						, (data->int_part % ten_E[int_digits])
						, (int)fra_digits
						, (data->fra_part / ten_E[MAX_Digits - fra_digits])
						);
		}
		else { /* integer digits aren't needed */
			(void)snprintf(str
						, SZ_Numeral_str
						, "%s.%0*llu"
						, ((IS_Negative(data) != 0)? "-": "")
						, (int)fra_digits
						, (data->fra_part / ten_E[MAX_Digits - fra_digits])
						);
		}
	}
	else { /* fractional digits aren't needed */
		(void)snprintf(str
						, SZ_Numeral_str
						, "%s%llu"
						, ((IS_Negative(data) != 0)? "-": "")
						, (data->int_part % ten_E[int_digits])
						);
	}

	return	str; /* normal end */
}


/* core of comparing two data structures */
static int cmp_data(Data_t *a, Data_t *b)
{
	int		result;
	#define	Lesser	-1
	#define	Equals	0
	#define	Greater	1

	if (((result = IS_Negative(a)) != 0) && IS_Negative(b) == 0) {
		return	Lesser; /* a is less than b. */
	}
	if (result == 0 && IS_Negative(b) != 0) {
		return	Greater;  /* a is greater than b. */
	}

	result	= (result != 0)? -1: 1; /* set signature coefficient */

	/* compare integer parts */
	if (a->int_part < b->int_part) {
		/* integer part of a is less than b's => abs(a) is less than abs(b)
		 */
		return	result * Lesser;
	}
	else if (a->int_part > b->int_part) {
		/* integer part of a is greater than b's
		 * => abs(a) is greater than abs(b)
		 */
		return	result * Greater;
	}

	/* integer parts are same => compare fraction part */
	return	result * (int)(a->fra_part - b->fra_part);
}


/* `Nop' opcode */
static long long nop(Machine_t *mac, void **operands)
{
	/* machine isn't defined? */
	if ((mac == NULL) || operands == NULL) {
		errno = EINVAL; return !NORMAL_END;
	}

	/* skip to the next command */
	++(mac->pc); return NORMAL_END;
}


/* `Ret' opcode */
static long long ret(Machine_t *mac, void **operands)
{
	/* machine isn't defined? */
	if ((mac == NULL) || operands == NULL) {
		errno = EINVAL; return !NORMAL_END;
	}

	/* return the specified value */
	return	(RET_CAST)(operands[0]);
}


/* `Set' opcode */
static long long set(Machine_t *mac, void **operands)
{
	Data_t	*d, *s;
	void	*dst = operands[0], *src = operands[1];
	char	sbuf[SZ_Numeral_str];

	/* test: source operand is same as destiny? */
	if ((d = (Data_t *)get_address(mac, dst))
			== (s = (Data_t *)get_address(mac, src))
	) {	++(mac->pc); return NORMAL_END; /* normal end */	}

	/* initialize sbuf[] */
	memset(sbuf, (int)'\0', sizeof(sbuf));

	/* test: destiny operand points to invoker's buffer? */
	if (d == dst) {
		/* destiny operand points to invoker's buffer */
		Data_t	x;

		/* test: source operand points to invoker's buffer? */
		if (s == src) {
			/* convert source string to data */
			if (str_to_data(retrieve_operand((char *)s, sbuf)
							, (Data_t *)memset(&x, (int)'\0', sizeof(x))
							) == NULL
			) {	return !NORMAL_END; /* ERROR: occurred several errors */ }

			s	= &x;
		}

		/* test: failed to convert source data to destiny string? */
		if (data_to_str(s, (char *)d, MAX_Digits, MAX_Digits) == NULL) {
			return	!NORMAL_END; /* ERROR: occurred several errors */
		}
	}
	else if (s == src) {
		/* source operand points to invoker's buffer
		 * => convert source string to data
		 */
		if (str_to_data(retrieve_operand((char *)s, sbuf), d) == NULL) {
			return	!NORMAL_END; /* ERROR: occurred several errors */
		}
	}
	else {	/* source and destiny operand points to register#0
			 * or working memory => copy source to destiny
			 */
		*d	= *s;
	}

	++(mac->pc); return NORMAL_END; /* normal end */
}


/* `Jmp' opcode */
static long long jmp(Machine_t *mac, void **operands)
{
	ssize_t	skip_steps = (ssize_t)(operands[0]);

	/* test: the skip step is illegal? */
	if (IS_Illegal_skip_steps(skip_steps, mac)) {
		errno = EINVAL; return !NORMAL_END; /* ERROR: invalid arguments */
	}

	/* compare the data and set `pc' according to the result */
	mac->pc	+= skip_steps;

	return	NORMAL_END; /* normal end */
}


/* `Cmp' opcode */
static long long cmp(Machine_t *mac, void **operands)
{
	ssize_t	less_steps	= (ssize_t)(operands[2])
			, equals_steps	= (ssize_t)(operands[3])
			, greater_steps	= (ssize_t)(operands[4])
			;
	Data_t	a, *ap, b, *bp;
	void	*val_a = operands[0], *val_b = operands[1];
	char	abuf[SZ_Numeral_str], bbuf[SZ_Numeral_str];
	int		cmp_result;

	/* test: the skip steps are illegal? */
	if (((IS_Illegal_skip_steps(less_steps, mac))
	|| IS_Illegal_skip_steps(equals_steps, mac))
	|| IS_Illegal_skip_steps(greater_steps, mac)
	) {
		errno = EINVAL; return !NORMAL_END; /* ERROR: invalid arguments */
	}

	/* test: the values are same. */
	if ((ap = (Data_t *)get_address(mac, val_a))
			== (bp = (Data_t *)get_address(mac, val_b))
	) {	cmp_result	= 0; /* a equals b. */	}

	/* if operand#0 indicates invoker's string buffer,
	 * convert the string to `a' data and set 'ap' pointer
	 */
	if (ap == val_a) {
		if (str_to_data(retrieve_operand((char *)ap
										, (char *)memset(abuf
														, (int)'\0'
														, sizeof(abuf)
														)
										)
						, (Data_t *)memset(&a, (int)'\0', sizeof(a))
						) == NULL
		) {	return	!NORMAL_END; /* ERROR: occurred several errors */	}

		ap	= &a;
	}

	/* if operand#1 indicates invoker's string buffer,
	 * convert the string to `b' data and set 'bp' pointer
	 */
	if (bp == val_b) {
		if (str_to_data(retrieve_operand((char *)bp
										, (char *)memset(bbuf
														, (int)'\0'
														, sizeof(bbuf)
														)
										)
						, (Data_t *)memset(&b, (int)'\0', sizeof(b))
						) == NULL
		) {	return	!NORMAL_END; /* ERROR: occurred several errors */	}

		bp	= &b;
	}

	/* compare the data and set `pc' according to the result */
	mac->pc += (((cmp_result = cmp_data(ap, bp)) == Lesser)? less_steps
					: ((cmp_result == Greater)?				greater_steps
					:										equals_steps
					));

	return	NORMAL_END; /* normal end */
}


/* core of adding the two data structures */
static long long add_data(Machine_t *mac, Data_t *ap, Data_t *bp)
{
	Data_t	*r0 = &(mac->r0);

	/* signature of result is same as signature of `Augend (ap)' */
	r0->is_negative	= ap->is_negative;

	/* test: fraction part of result is greater than the maximum? */
	if ((r0->fra_part = ap->fra_part + bp->fra_part) > MAX_Fra_part) {
		/* carry up integer part of result */
		r0->int_part	= r0->fra_part / ten_E[MAX_Exp];
		r0->fra_part	%= ten_E[MAX_Exp];
	}
	else {	/* initialize integer part of result */
		r0->int_part	= 0ULL;
	}

	/* test: integer part of `Augend (ap)' equals the maximum, or
	 *	integer part of result will be greater than the maximum?
	 */
	if ((ap->int_part == MAX_Int_part)
	|| bp->int_part > (ten_E[MAX_Exp] - ap->int_part - r0->int_part)
	) {	/* ERROR: result is out of range */
		errno = ERANGE; return !NORMAL_END;
	}

	/* set integer part of result */
	r0->int_part	+= (ap->int_part + bp->int_part);

	return	NORMAL_END; /* normal end */
}


/* core of subtracting the two data structures */
static long long sub_data(Machine_t *mac, Data_t *ap, Data_t *bp)
{
	Data_t	*r0 = &(mac->r0);
	int	sign_of_a = ap->is_negative; /* save signature of `Minuend' */

	/* test: signatures of `Minuend' and `Subtrahend' aren't equal? */
	if (sign_of_a != bp->is_negative) {
		/* in this case, the result should be
		 *		<signatures of `Minuend'> * (|`Minuend'| + |`Subtrahend'|)
		 */
		bp->is_negative	= sign_of_a;
		return	add_data(mac, ap, bp);
	}

	/* set signature of result: (1) get absolutes of values */
	ap->is_negative = bp->is_negative = 0;

	/* (2) test: `Minuend' is least than `Subtrahend'? */
	if (cmp_data(ap, bp) < 0) {
		/* (2)-1 swap the `Minuend' and `Subtrahend', and set signature of
		 * result to the inverted signature of the `Minuend'
		 */
		Data_t *t = ap; ap = bp; bp = t;
		r0->is_negative	= ((sign_of_a == 0)? 1: sign_of_a);
	}
	else {	/* (2)-2 set signature of result same as `Minuend' signature */
		r0->is_negative= ap->is_negative;
	}

	/* (3) set signatures of values same as `Minuend' */
	ap->is_negative = bp->is_negative = sign_of_a;

	/* test: fraction part of `Minuend' is least than `Subtrahend'?
	 */
	if (ap->fra_part < bp->fra_part) {
		/* borrow the integer part of `Minuend',
		 * and set fraction part of result
		 */
		--ap->int_part;
		r0->fra_part =	(ten_E[MAX_Exp] - bp->fra_part) + ap->fra_part;
	}
	else {	/* set fraction part of result */
		r0->fra_part =	ap->fra_part - bp->fra_part;
	}

	/* set integer part of result */
	r0->int_part	= ap->int_part - bp->int_part;

	return	NORMAL_END; /* normal end */
}


/* `Sub' opcode */
static long long sub(Machine_t *mac, void **operands)
{
	Data_t	minuend, subtrahend, *ap, *bp = NULL;
	int		result;

	/* test: the operands are illegal?
	 * (if those are legal, translate operands to data)
	 */
	if (((ap = operand_to_data(mac
							, operands[0]
							, (Data_t *)memset(&minuend
												, (int)'\0'
												, sizeof(minuend)
												)
							)) == NULL)
	|| (bp = operand_to_data(mac
							, operands[1]
							, (Data_t *)memset(&subtrahend
												, (int) '\0'
												, sizeof(subtrahend)
												)
							)) == NULL
	) {	return	!NORMAL_END; /* ERROR: occurred several errors */	}

	/* test: subtracting of data had executed normally? */
	if ((result = sub_data(mac, ap, bp)) == NORMAL_END) {
		++(mac->pc);	/* go to next command */
	}

	return	result;	/* result of subtracting */
}


/* `Add' opcode */
static long long add(Machine_t *mac, void **operands)
{
	Data_t	augend, addend, *ap, *bp = NULL;
	int		result;

	/* test: the operands are illegal?
	 * (if those are legal, translate operands to data)
	 */
	if (((ap = operand_to_data(mac
								, operands[0]
								, (Data_t *)memset(&augend
													, (int)'\0'
													, sizeof(augend)
													)
								)) == NULL)
	|| (bp = operand_to_data(mac
							, operands[1]
							, (Data_t *)memset(&addend
												, (int)'\0'
												, sizeof(addend)
												)
							)) == NULL
	) {	return	!NORMAL_END; /* ERROR: occurred several errors */	}

	/* test: signatures of `Augend' and `Addend'... */
	if (ap->is_negative == bp->is_negative) {
		/* signatures are equal => execute adding data */
		result	= add_data(mac, ap, bp);
	}
	else if (IS_Negative(ap) != 0) {
		/* signatures are unequal and `Augend' is negative
		 * => execute "`Augend' - |`Addend'|"
		 */
		ap->is_negative	= 0;
		result	= sub_data(mac, bp, ap);
	}
	else {	/* signatures are unequal and `Augend' is negative
			 * => execute "`Augend' - |`Addend'|"
			 */
		bp->is_negative	= 0;
		result	= sub_data(mac, ap, bp);
	}

	/* test: adding or subtracting had executed normally? */
	if (result == NORMAL_END) {
		++(mac->pc);	/* go to next command */
	}

	return result;	/* normal end */
}


/* core of multiplying the two data structures */
static long long mul_data(Machine_t *mac, Data_t *ap, Data_t *bp)
{
	unsigned long long	a[ELEM_N_Multiply_vector]
						, b[ELEM_N_Multiply_vector]
						, r[ELEM_N_Multiply_vector]
						, digits_of_elem = ten_E[MAX_Exp / 2]
						;
	Data_t				*r0 = &(mac->r0);
	size_t				i, j, k;

	/* resolve operands to `a' and `b' vectors */
	a[0]	= ap->int_part / digits_of_elem;
	a[1]	= ap->int_part % digits_of_elem;
	a[2]	= ap->fra_part / digits_of_elem;
	a[3]	= ap->fra_part % digits_of_elem;
	b[0]	= bp->int_part / digits_of_elem;
	b[1]	= bp->int_part % digits_of_elem;
	b[2]	= bp->fra_part / digits_of_elem;
	b[3]	= bp->fra_part % digits_of_elem;

	/* test: the 10E elements of a-vector and b-vector are not zero?
	 */
	if ((a[0] != 0ULL) &&  b[0] != 0ULL) {
		/* ERROR: result should be out of range */
		errno	= ERANGE; return !NORMAL_END;
	}

	/* calculate `r' vector from a-vector and b-vector
	 * (1) make r[0] to be zero
	 * (1) calculate 10E-18 element of r-vector
	 * (2) calculate 10E-9, 10E+0 and 10E+9 elements of r-vector
	 */
	memset(r, (int)'\0', sizeof(r[0]));
	r[ELEM_N_Multiply_vector - 1] = (
			(a[3] * b[1] + a[2] * b[2] + a[1] * b[3])
				+ (((a[3] * b[2] + a[2] * b[3])
					+ ((a[3] * b[3]) / digits_of_elem)
					) / digits_of_elem)
			);
	for (i = ELEM_N_Multiply_vector - 1
	; i-- > 0
	; r[i + 1] %= digits_of_elem
	) {
		for (j = i + 2, k = 0, r[i] = r[i + 1] / digits_of_elem
		; j-- > 0
		; ++k
		) {	r[i]	+= (a[j] * b[k]);	}
	}

	/* test: 10E+9 elements of r-vector is equals or greater than 10E+9? */
	if (r[0] >= digits_of_elem) {
		/* ERROR: the result should be out of range */
		errno	= ERANGE; return !NORMAL_END;
	}

	/* set integrated r-vector to result */
	r0->int_part	= digits_of_elem * r[0] + r[1];
	r0->fra_part	= digits_of_elem * r[2] + r[3];

	/* set signature of result */
	r0->is_negative	= ((IS_Zero(r0) || ap->is_negative == bp->is_negative)?
																	0: !0);

	return	NORMAL_END; /* normal end */
}


/* `Mul' opcode */
static long long mul(Machine_t *mac, void **operands)
{
	Data_t	multiplicand, multiplier, *ap, *bp = NULL;
	int		result;

	/* test: the operands are illegal?
	 *	(if those are legal, translate operands to data)
	 */
	if (((ap = operand_to_data(mac
								, operands[0]
								, (Data_t *)memset(&multiplicand
													, (int)'\0'
													, sizeof(multiplicand)
													)
								)) == NULL)
	|| (bp = operand_to_data(mac
							, operands[1]
							, (Data_t *)memset(&multiplier
												, (int)'\0'
												, sizeof(multiplier)
												)
							)) == NULL
	) {	return	!NORMAL_END; /* ERROR: occurred several errors */	}

	/* calculate result of multiplies operand#0 and operand#1 */
	result	= mul_data(mac, ap, bp);

	/* test: multiplying of data had executed normally? */
	if (result == NORMAL_END) {
		++(mac->pc);	/* go to next command */
	}

	return result;	/* normal end */
}


/* core of dividing numerator by denominator and calculating remainder
 */
static long long divide_data(Machine_t *mac
						, Data_t *ap
						, Data_t *bp
						, Data_t *remp
						)
{
	unsigned long long	n, denominator;
	Data_t				*r0 = &(mac->r0); /* register#0 */
	size_t				i, int_digits, fra_digits = 0ULL;
	char				numerator[SZ_Numerator], *np, sv_ch;

	/* test: invoker divides by zero? */
	if (IS_Zero(bp)) {
		/* ERROR: argument is out of domain */
		errno = EDOM; return !NORMAL_END;
	}

	/* calculate denominator as integer-type, and count up digits
	 * of fraction-part of it
	 */
	denominator	= bp->int_part;
	if (bp->fra_part > 0ULL) {
		for (i = MAX_Digits, n = bp->fra_part; i-- > 0; ) {
			if ((unsigned long long)(LLONG_MAX / 10LL) < denominator) {
				/* ERROR: argument is out of domain */
				errno = EDOM; return !NORMAL_END;
			}

			denominator	= (denominator * 10ULL) + (n / ten_E[i]);

			/* test: the left digits are zero? */
			if ((n %= ten_E[i]) == 0ULL) {
				fra_digits	= MAX_Digits - i;
				break; /* stop this iteration */
			}
		}
	}

	/* count up digits of integer-part of denominator */
	for (i = MAX_Digits, int_digits = 0; i-- > 0; ) {
		if (bp->int_part >= ten_E[i]) {
			int_digits	= i + 1;
			break; /* stop this iteration */
		}
	}

	/* make numerator to array of digit-characters */
	#define Iteration { /* convert digit to character */	\
		*np	= i2a[n / ten_E[i]]; n %= ten_E[i];	\
	}
	np	= numerator;
	for (i = MAX_Digits, n = ap->int_part; i-- > 0; ++np) {
		Iteration; /* make integer digits to be characters */
	}
	for (i = MAX_Digits, n = ap->fra_part; i-- > 0; ++np) {
		Iteration; /* make fractional digits to be characters */
	}
	memset(np, (int)'0', fra_digits); /* append `fra_digits' zero codes */
	*(np + fra_digits) = '\0'; /* put terminate code */
	#undef Iteration

	/* calculate result to divide of numerator by denominator */
	#define Iteration(PART) { /* calculation core */			\
		sv_ch	= *(++np); *np = '\0';							\
		n	= (n * 10ULL) + atoi(np - 1);						\
		*np	= sv_ch;											\
		if (r0->PART != 0ULL) {									\
			r0->PART	*= 10ULL;								\
		}														\
		if (n >= denominator) {									\
			r0->PART	+= (n / denominator);					\
		}														\
		if (n != 0ULL) {										\
			n	= (unsigned long long)(							\
						(long long)n % (long long)denominator	\
					);											\
		}														\
	}

	for (i = 0, r0->int_part = 0ULL, np = numerator, n = 0ULL
		, int_digits = MAX_Digits + fra_digits
	; i < int_digits
	; ++i
	) {
		Iteration(int_part); /* calculate integer part */

		if (r0->int_part > MAX_Int_part) {
			errno = ERANGE; return !NORMAL_END;
		}
	}
	for (i = 0, r0->fra_part = 0ULL; i < MAX_Digits; ++i) {
		Iteration(fra_part); /* calculate fractional part */
	}
	#undef Iteration

	/* set signature of result */
	r0->is_negative	= (IS_Zero(r0) || ap->is_negative == bp->is_negative)?
																	0: !0;

	/* test: invoker needs remainder, too? */
	if (remp != NULL) {
		Data_t	mul_res, sv_r0 = *r0; /* save register#0 */

		/* calculate remainder */
		(void)mul_data(mac, &sv_r0, bp);
		mul_res	= *r0;
		(void)sub_data(mac, ap, &mul_res);

		/* test: destination of remainder isn't register#0? */
		if (remp != r0) {
			/* destination of remainder is a memory-slot */
			*remp	= *r0;		/* transfer remainder to memory */
			*r0		= sv_r0;	/* restore register#0 */
		}

		/* set signature of remainder */
		remp->is_negative = (IS_Zero(remp) != 0 || IS_Negative(ap) == 0)?
																	0: !0;
	}

	return	NORMAL_END; /* normal end */
}


/* `Div' opcode */
static long long divide(Machine_t *mac, void **operands)
{
	Data_t	dividend, divisor, *remp, *ap, *bp = NULL;
	int		result;

	/* test: the operand#0 and operand#1 are illegal?
	 *	(if those are legal, translate operands to data)
	 */
	if (((ap = operand_to_data(mac
								, operands[0]
								, (Data_t *)memset(&dividend
													, (int)'\0'
													, sizeof(dividend)
													)
								)) == NULL)
	|| (bp = operand_to_data(mac
								, operands[1]
								, (Data_t *)memset(&divisor
													, (int)'\0'
													, sizeof(divisor)
													)
								)) == NULL
	) {	return !NORMAL_END; /* ERROR: occurred several errors */	}

	/* test: which destination of remainder does the operand#2 indicate? */
	if (operands[2] == NO_Remainder) {
		remp	= NULL; /* invoker needs no remainder */
	}
	else if (operands[2] == Register0) {
		remp	= &(mac->r0); /* invoker needs remainder only */
	}
	else if ((mac->memories_n > 0)
	&& (size_t)(operands[2]) < mac->memories_n
	) {	/* invoker needs to set remainder to a memory slot */
		remp	= mac->memories + (size_t)(operands[2]);
	}
	else {	
		/* ERROR: invoker needs to set remainder to a string */
		errno = EINVAL; return !NORMAL_END;
	}

	/* calculate result of divide operand#0 by operand#1 */
	result	= divide_data(mac, ap, bp, remp);

	/* test: dividing of data had executed normally? */
	if (result == NORMAL_END) {
		++(mac->pc);	/* go to next command */
	}

	return result;	/* normal end */
}


static void rounding_data(unsigned long long	*partp
						, ssize_t				exp
						, How_to_rounding_e		how 
						)
{
	unsigned long long	x = (*partp / ten_E[exp + 1]) * ten_E[exp + 1];

	/* test: invoker needs rounding down? */
	if (how == ROUND_down) {
		*partp	= x; /* set result */
	}
	else {	/* invoker needs rounding even or up */
		unsigned long long	y = (*partp % ten_E[exp + 1]) / ten_E[exp];

		*partp	= x + (((how == ROUND_up && y > 0) || y >= 5)?
														ten_E[exp + 1]: 0
						); /* set result */
	}

	return; /* return value is nothing */
}


static long long rounding(Machine_t *mac, void **operands)
{
	How_to_rounding_e	how = (How_to_rounding_e)(operands[2]);
	ssize_t				exp = (ssize_t)(operands[1]);
	Data_t				*r0;

	/* test: the operand#0, #1, #2 are illegal? */
	if ((((((r0 = operand_to_data(mac, operands[0], &(mac->r0))) == NULL)
	|| how < ROUND_up) || ROUND_down < how) 
	|| exp < -((ssize_t)(MAX_Exp)))
	|| (ssize_t)(MAX_Exp - 1) < exp
	) {	/* ERROR: occurred several errors */
		errno = EINVAL; return !NORMAL_END;
	}

	/* test: exponent is negative? */
	if (exp < 0) {
		/* rounding digit indicates fractional part */
		rounding_data(&(r0->fra_part), MAX_Exp + exp, how);

		/* test: occurred `carry up'? */
		if (r0->fra_part > MAX_Fra_part) {
			r0->fra_part	%= ten_E[MAX_Exp];
			++(r0->int_part); /* carry up integer part */
		}
	}
	else {	/* rounding digit indicates integer part */
		rounding_data(&(r0->int_part), exp, how);
		r0->fra_part	= 0ULL; /* fractional must be zero */
	}

	++(mac->pc); /* skip to the next step */

	return NORMAL_END; /* normal end */
}


/* definition of abacus architecture */
int abacus(char		*result
		, size_t	result_sz
		, size_t	int_digits
		, size_t	fra_digits
		, size_t	memories_n
		, size_t	commands_n
		, Command_t	*commands
		, size_t	*pc
		)
{	/* define the pairs of opcodes and operator-functions */
	static struct opcode_def_t {
		Opcode_e	opcode;
		long long	(*operator)(Machine_t *mac, void **operands);
	}							opcode_defs[] = { { Nop_op, nop			}
												, { Ret_op, ret			}
												, { Set_op, set			}
												, { Jmp_op, jmp			}
												, { Cmp_op, cmp			}
												, { Sub_op, sub			}
												, { Add_op, add			}
												, { Mul_op, mul			}
												, { Div_op, divide		}
												, { Rou_op, rounding	}
												};
	Command_t					*cmd;
	Machine_t					mac;
	long long					retcode;

	/* test: arguments are invalid? */
	if (((((((result == NULL) || commands == NULL) || pc == NULL)
	|| int_digits > MAX_Digits) || fra_digits > MAX_Digits)
	|| (int_digits == 0 && fra_digits == 0))
	|| result_sz < (1 + int_digits + 1 + fra_digits + 1)
	) {	return !NORMAL_END; /* ERROR: invalid arguments */	}

	/* test: failed to create the machine? */
	if (create_machine((Machine_t *)memset(&mac, (int)'\0', sizeof(mac))
						, memories_n
						, commands_n
						) == NULL
	) {	/* ERROR: occurred several errors */
		(void)destroy_machine(&mac);
		return	!NORMAL_END;
	}

	/* execute the calculations... */
	*pc = 0; retcode = NORMAL_END;
	while ((retcode == NORMAL_END) && mac.pc < commands_n) {
		/* look up the `opcode_defs[]'.  if the opcode is unknown,
		 * this API returns error.
		 */
		cmd	= commands + mac.pc;
		if (cmd->opcode >= (sizeof(opcode_defs) / sizeof(opcode_defs[0]))
		) {	/* ERROR: invalid arguments (unknown opcode) */
			errno = EINVAL; retcode = !NORMAL_END; break;
		}

		/* execute the operator, and progress the pc */
		if (((retcode = (*(opcode_defs[cmd->opcode].operator))(&mac
														, cmd->operands
														)) != NORMAL_END)
		|| cmd->opcode == Ret_op
		) {	break; /* ERROR: occurred several errors */	}
	}

	/* output program counter */
	*pc	= mac.pc;

	/* test: failed to convert the register#0 to result?
	 * or failed to destroy the machine?
	 */
	if (data_to_str(&(mac.r0), result, int_digits, fra_digits) == NULL) {
		retcode	= !NORMAL_END; /* ERROR: occurred several errors */
	}
	if (destroy_machine(&mac) == NULL) {
		retcode	= !NORMAL_END; /* ERROR: occurred several errors */
	}

	return	(int)retcode; /* normal end/occurred errors */
}
